//
// zi80.cpp -- Z-80, 8080 and 8085 emulator
//
// Completeness of support decreases in that order.
//

#include <assert.h>

#include "zi80.h"

/* Based on refz80.c, of course. Going back to trs83. */

/* I've gone one step past Peter's work on this to catch all the z80 memory
 * writes that happen through pointers.
 */

/*
 * Bus activity is accurate for the core instructions, but may be off for
 * shift/rotate (CB) group and IX/IY.  Some of the latter have been fixed up.
 * By extension, the DD|FD CB shift/rotate on IX/IY are quite suspect.
 */

void Zi80::PowerUp()
{
	m_reg.PC = 0;
	m_cycles = 0;
	m_tcount = 0;

	m_reg.AF = 0;
	m_reg.BC = 0;
	m_reg.DE = 0;
	m_reg.HL = 0;
	m_reg.AFp = 0;
	m_reg.BCp = 0;
	m_reg.DEp = 0;
	m_reg.HLp = 0;
	m_reg.IX = 0xffff;
	m_reg.IY = 0;
	m_reg.I = 0;
	m_reg.IFF1 = 0;
	m_reg.IFF2 = 0;
	m_reg.IM = 0;
	m_reg.R = 0;
	m_reg.SP = 0;

	m_INT = 0;

	m_rcounter = 0;
}

// Base implemenations of bus.  Accurate because you'll get 0xff because
// of the floating value with nothing attached.

u8		Zi80::OpcodeFetch(u16 address)			{ return 0xff; }
u8		Zi80::IntrFetch()						{ return 0xff; }
u8		Zi80::DataRead(u16 address)				{ return 0xff; }
void	Zi80::DataWrite(u16 address, u8 data)	{			   };
u8		Zi80::PortRead(u16 address)				{ return 0xff; }
void	Zi80::PortWrite(u16 address, u8 data)	{			   };

/* values which hide little/big endian differences */
static int lowoff;
static int highoff;

#ifdef HOST_LITTLE_ENDIAN
#define low(w)	((u8*)&(w) + 0)
#define high(w)	((u8*)&(w) + 1)
#endif

#ifdef HOST_BIG_ENDIAN
#define low(w)	((u8*)&(w) + 1)
#define high(w)	((u8*)&(w) + 0)
#endif

#ifndef HOST_BIG_ENDIAN
#ifndef HOST_LITTLE_ENDIAN
#define low(w)	((u8*)&(w) + lowoff)
#define high(w)	((u8*)&(w) + highoff)
#endif
#endif

#define a	high(m_reg.AF)
#define f	low(m_reg.AF)
#define b	high(m_reg.BC)
#define c	low(m_reg.BC)
#define d	high(m_reg.DE)
#define e	low(m_reg.DE)
#define h	high(m_reg.HL)
#define l	low(m_reg.HL)

#define af	m_reg.AF
#define bc	m_reg.BC
#define de	m_reg.DE
#define hl	m_reg.HL
#define ix	m_reg.IX
#define iy	m_reg.IY
#define pc	m_reg.PC
#define sp	m_reg.SP
#define afp	m_reg.AFp
#define bcp	m_reg.BCp
#define dep	m_reg.DEp
#define hlp	m_reg.HLp

#define carry	(*f & 1)

#define cyc( N ) m_cycles -= N

// Silly, but placeholder reminder for  utility class.
#define BUS_INST_READ	(0)
#define BUS_READ		(1)
#define BUS_WRITE		(2)

void Zi80::bus_op(int type, u16 addr, i32 cycles, u8 data)
{
	m_tcount += cycles;
}

// XXX - anything which fetches instructions or instruction data
// should check the interrupt line and get data from external interrupt
// control via IntrFetch().  But, practically speaking, we're unlikely
// to run into anything that does more than RST instructions.

u8 Zi80::fetch_opcode()
{
	bus_op(BUS_INST_READ, pc, 4, 0);

	u8 op = OpcodeFetch(pc);
	pc++;

	m_rcounter++;

	return op;
}


u8 Zi80::fetch_n()
{
	bus_op(BUS_READ, pc, 3, 0);

	u8 data = DataRead(pc);
	pc++;

	return data;
}

u16 Zi80::fetch_nn()
{
	bus_op(BUS_READ, pc, 3, 0);
	bus_op(BUS_READ, pc + 1, 3, 0);

    u8 lb = DataRead(pc);
	pc++;
    u8 hb = DataRead(pc);
	pc++;

    return (hb << 8) | lb;
}

void Zi80::store8(u16 mem, u8 val)
{
	bus_op(BUS_WRITE, mem, 3, val);
	DataWrite(mem, val);
}

void Zi80::store16(u16 mem, u16 val)
{
	bus_op(BUS_WRITE, mem, 3, (u8)(val & 0xff));
	bus_op(BUS_WRITE, mem + 1, 3, (u8)((val >> 8) & 0xff));

	DataWrite(mem, (u8)(val & 0xff));
	DataWrite((u16)(mem + 1), (u8)((val >> 8) & 0xff));
}

u8 Zi80::load8(u16 mem)
{
	bus_op(BUS_READ, mem, 3, 0);

	return DataRead(mem);
}

u16 Zi80::load16(u16 mem)
{
	bus_op(BUS_READ, mem, 3, 0);
	bus_op(BUS_READ, mem + 1, 3, 0);

    u8 lb = DataRead(mem);
	mem++;
    u8 hb = DataRead(mem);
	mem++;

    return (hb << 8) | lb;
}

u16 Zi80::ixy(u16 *xy)
{
	bus_op(BUS_READ, pc, 3, 0);
	m_tcount += 5;	/* compute IX+d; costly! */

	u8 val = *xy + (i8)DataRead(pc);

	pc++;

	return val;
}

/* utility routines for dealing with Z80 flag register */

void Zi80::assign_s(int val)
{
    *f &= ~0x80;
    if (val) *f |= 0x80;
}

void Zi80::assign_z(int val)
{
    *f &= ~0x40;
    if (val) *f |= 0x40;
}

void Zi80::assign_h(int val)
{
    *f &= ~0x10;
    if (val) *f |= 0x10;
}

void Zi80::assign_pv(int val)
{
    *f &= ~4;
    if (val) *f |= 4;
}

void Zi80::assign_n(int val)
{
    *f &= ~2;
    if (val) *f |= 2;
}

void Zi80::assign_c(int val)
{
    *f &= ~1;
    if (val) *f |= 1;
}

#define if_s	(*f & 0x80)
#define if_z	(*f & 0x40)
#define if_pv	(*f & 0x04)
#define if_c	(*f & 0x01)

static int parity(u8 val)
{
    int cnt;

    cnt =
		!!(val & 0x80) +
		!!(val & 0x40) +
		!!(val & 0x20) +
		!!(val & 0x10) +
		!!(val & 0x08) +
		!!(val & 0x04) +
		!!(val & 0x02) +
		!!(val & 0x01);

    return !(cnt & 1);
}

/* routines which do most of the opcode implementation */

void Zi80::add(u8 *acc, int val, int cry, int docarry)
{
    int res;

    res = (int)*acc + val + cry;
    assign_s(res & 0x80);
    if (docarry) assign_c(res & 0x100);
    assign_h(((*acc & 15) + (val & 15) + cry) & 16);
    assign_n(0);

    assign_z(!(res & 0xff));

    res = (i8)*acc + (i8)val + cry;
    assign_pv(res > 127 || res < -128);

    *acc = (int)*acc + val + cry;
}

void Zi80::sub(u8 *acc, int val, int cry, int docarry)
{
    int res;

    res = (int)*acc - val - cry;
    assign_s(res & 0x80);
    if (docarry) assign_c(res & 0x100);
    assign_h(((*acc & 15) - (val & 15) - cry) & 16);
    assign_n(1);

    assign_z(!(res & 0xff));

    res = (i8)*acc - (i8)val - cry;
    assign_pv(res > 127 || res < -128);

    *acc = (int)*acc - val - cry;
}

void Zi80::cp(int val)
{
    u8 dummy = *a;
    sub(&dummy, val, 0, 1);
}

#define inc(r) add(r, 1, 0, 0)
#define dec(r) sub(r, 1, 0, 0)

void Zi80::andz(int val)
{
    *a &= val;
    assign_s(*a & 0x80);
    assign_z(!*a);
    assign_h(1);
    assign_pv(parity(*a));
    assign_n(0);
    assign_c(0);
}

void Zi80::orz(int val)
{
    *a |= val;
    assign_s(*a & 0x80);
    assign_z(!*a);
    assign_h(0);
    assign_pv(parity(*a));
    assign_n(0);
    assign_c(0);
}

void Zi80::xorz(int val)
{
    *a ^= val;
    assign_s(*a & 0x80);
    assign_z(!*a);
    assign_h(0);
    assign_pv(parity(*a));
    assign_n(0);
    assign_c(0);
}

static void ex(u16 *x, u16 *y)
{
    u16 temp;
    temp = *x;
    *x = *y;
    *y = temp;
}

void Zi80::ex_memsp(u16 *r)
{
    u16 temp;

	// XXX - wrong; the byte should be interleaved.
    temp = *r;
    *r = load16(sp);
    store16(sp, temp);
}

void Zi80::add16(u16 *acc, int val)
{
    int r;

    assign_n(0);
    assign_h(((*acc & 0xfff) + (val & 0xfff)) & 0x1000);
    /* Turbo C fucks this up; do it the hard way */
    /*res = (long)*acc + (long)val;*/
    r = (*low(*acc) + *low(val)) >> 8;
    assign_c((*high(*acc) + *high(val) + r) & 0x100);
    *acc += val;
}

void Zi80::adc16(u16 *acc, int val)
{
    add((u8*)acc + lowoff, val & 0xff, carry, 1);
    add((u8*)acc + highoff, (val >> 8) & 0xff, carry, 1);
    assign_z(!*acc);
}

void Zi80::sbc16(u16 *acc, int val)
{
    sub((u8*)acc + lowoff, val & 0xff, carry, 1);
    sub((u8*)acc + highoff, (val >> 8) & 0xff, carry, 1);
    assign_z(!*acc);
}

void Zi80::daa()
{
    int res = *a;

    if (*f & 2) {
		if ((res & 15) > 9 || (*f & 16))
		    res -= 6;
		if (*a > 0x99 || carry)
		    res -= 0x60;
		assign_h((*f & 16) && (*a & 15) < 6);
		if (*a > 0x99)
		    assign_c(1);
		*a = res;
    }
    else {
		if ((res & 15) > 9 || (*f & 16))
		    res += 6;
		if (res > 0x9f || carry)
		    res += 0x60;
		if (res > 255)
		    assign_c(1);
		assign_h((*a & 15) > 9);
		*a = res;
    }

    assign_s(*a & 0x80);
    assign_z(!*a);
    assign_pv(parity(*a));
}


void Zi80::rs_flags(u8 val)
{
    assign_z(!val);
    assign_s(val & 0x80);
    assign_h(0);
    assign_pv(parity(val));
    assign_n(0);
}

void Zi80::rlca()
{
    assign_h(0);
    assign_n(0);
    assign_c(*a & 0x80);
    *a <<= 1;
    *a |= carry;
}

void Zi80::rla()
{
    int prevc = carry;

    assign_h(0);
    assign_n(0);
    assign_c(*a & 0x80);
    *a <<= 1;
    *a |= prevc;
}

void Zi80::rrca()
{
    assign_h(0);
    assign_n(0);
    assign_c(*a & 1);
    *a = (*a >> 1) & 0x7f;
    *a |= carry << 7;
}

void Zi80::rra()
{
    int prevc = carry;

    assign_h(0);
    assign_n(0);
    assign_c(*a & 1);
    *a = (*a >> 1) & 0x7f;
    *a |= prevc << 7;
}

void Zi80::rlc(u8 *acc)
{
    assign_c(*acc & 0x80);
    *acc <<= 1;
    *acc |= carry;
    rs_flags(*acc);
}

void Zi80::rl(u8 *acc)
{
    int prevc = carry;

    assign_c(*acc & 0x80);
    *acc <<= 1;
    *acc |= prevc;
    rs_flags(*acc);
}

void Zi80::rrc(u8 *acc)
{
    assign_c(*acc & 1);
    /* note: *acc >>= 1 should work, but I avoid it because some C
     * compilers screw it up.
     */
    *acc = (*acc >> 1) & 0x7f;
    *acc |= carry << 7;
    rs_flags(*acc);
}

void Zi80::rr(u8 *acc)
{
    int prevc = carry;

    assign_c(*acc & 1);
    *acc = (*acc >> 1) & 0x7f;
    *acc |= prevc << 7;
    rs_flags(*acc);
}

void Zi80::sla(u8 *acc)
{
    assign_c(*acc & 0x80);
    *acc <<= 1;
    rs_flags(*acc);
}

void Zi80::sra(u8 *acc)
{
    assign_c(*acc & 1);
    /* again, I don't trust right shifts */
    *acc = (*acc >> 1) & 0x7f;
    *acc |= (*acc << 1) & 0x80;
    rs_flags(*acc);
}

void Zi80::srl(u8 *acc)
{
    assign_c(*acc & 1);
    *acc = (*acc >> 1) & 0x7f;
    rs_flags(*acc);
}

void Zi80::bit(u8 val, int bit)
{
    assign_h(1);
    assign_n(0);
    assign_z(!(val & (1 << bit)));
}

void Zi80::set(u8 *val, int bit)
{
    *val |= 1 << bit;
}

void Zi80::res(u8 *val, int bit)
{
    *val &= ~(1 << bit);
}

void Zi80::pop(u16 *val)
{
	bus_op(BUS_READ, sp, 3, 0);
    *((u8*)val + lowoff) = DataRead(sp);
    sp++;
	bus_op(BUS_READ, sp, 3, 0);
    *((u8*)val + highoff) = DataRead(sp);
    sp++;
}

void Zi80::push(u16 val)
{
	m_tcount++;
    sp--;
	bus_op(BUS_WRITE, sp, 3, (u8)((val >> 8) & 255));
    DataWrite(sp, (u8)((val >> 8) & 255));
    sp--;
	bus_op(BUS_WRITE, sp, 3, (u8)(val & 255));
    DataWrite(sp, (u8)(val & 255));
}

bool Zi80::mode_1_IRQ()
{
	if (m_reg.IFF1 == 0)
		return false;

	m_tcount += 7 - 1;	// RST + 2, I believe.  The "-1" accounts for an extra in push()
	push(pc);

	pc = 0x38;
	m_reg.IFF1 = m_reg.IFF2 = 0;

	cyc(11 + 2);

	return true;
}

void Zi80::jump(int cond)
{
    u16 newpc = fetch_nn();

    if (cond) { pc = newpc; }
}

i32 Zi80::branch(int cond)
{
    u16 newpc;
	
	bus_op(BUS_READ, pc, 3, 0);

	newpc = (i8)DataRead(pc);
	pc++;

    newpc += pc;

    if (cond) { pc = newpc; return 12; }
	return 7;
}

void Zi80::call(int cond)
{
    u16 newpc = fetch_nn();

    if (cond) {
		push(pc);
		pc = newpc;
		cyc(7);		/* 7 extra on 10 makes 17 if condition true */
    }
}

void Zi80::in(u8 port, u8* acc, int doflags)
{
    *acc = PortRead(port);
    if (doflags) {
		assign_s(*acc & 0x80);
		assign_z(!*acc);
		assign_h(0);
		assign_pv(parity(*acc));
		assign_n(0);
    }
}

void Zi80::out(u8 port, u8 val)
{
	PortWrite(port, val);
}

Zi80::Zi80()
{
    u16 x;

    x = 0x0100;
    lowoff = *(u8*)&x;
    x = 0x0001;
    highoff = *(u8*)&x;

#ifdef HOST_BIG_ENDIAN
	assert(lowoff == 1);
#endif

#ifdef HOST_LITTLE_ENDIAN
	assert(lowoff == 0);
#endif

	PowerUp();
}

void Zi80::SetINT(u8 level)
{
	m_INT = level;
}

int Zi80::Step()
{
	i32 cyc0 = m_cycles;
	i32 tcount0 = m_tcount;
	int ret;

	u8 op;

	if (m_INT && m_reg.IFF1 == 1)
	{
		op = IntrFetch();
		m_cycles++;	// I believe

		m_reg.IFF1 = m_reg.IFF2 = 0;
	}
	else
	{
		op = fetch_opcode();
	}

    /* big switches can be hard on C compilers so I break things up */
    switch (op & 0xc0) {
    case 0x00: ret = main00(op); break;
    case 0x40: ret = main01(op); break;
    case 0x80: ret = main10(op); break;
    case 0xc0: ret = main11(op); break;
    }

	m_tcount = tcount0 + (cyc0 - m_cycles);

	return ret;
}

int Zi80::main00(u8 op)
{
    u8 t;

    switch (op) {
    case 0x00: cyc(4);  return 0;
    case 0x01: cyc(10); bc = fetch_nn(); return 0;
    case 0x02: cyc(7);  store8(bc, *a); return MEM_CHG;
    case 0x03: cyc(6);  bc++; return 0;
    case 0x04: cyc(4);  inc(b); return 0;
    case 0x05: cyc(4);  dec(b); return 0;
    case 0x06: cyc(7);  *b = fetch_n(); return 0;
    case 0x07: cyc(4);  rlca(); return 0;
    case 0x08: cyc(4);  ex(&af, &afp); return 0;
    case 0x09: cyc(11); add16(&hl, bc); return 0;
    case 0x0a: cyc(7);  *a = load8(bc); return 0;
    case 0x0b: cyc(6);  bc--; return 0;
    case 0x0c: cyc(4);  inc(c); return 0;
    case 0x0d: cyc(4);  dec(c); return 0;
    case 0x0e: cyc(7);  *c = fetch_n(); return 0;
    case 0x0f: cyc(4);  rrca(); return 0;
    case 0x10: (*b)--; cyc( branch(*b != 0) + 1); return 0;
    case 0x11: cyc(10); de = fetch_nn(); return 0;
    case 0x12: cyc(7);  store8(de, *a); return MEM_CHG;
    case 0x13: cyc(6);  de++; return 0;
    case 0x14: cyc(4);  inc(d); return 0;
    case 0x15: cyc(4);  dec(d); return 0;
    case 0x16: cyc(7);  *d = fetch_n(); return 0;
    case 0x17: cyc(4);  rla(); return 0;
    case 0x18: cyc(	branch(1) ); return 0;
    case 0x19: cyc(11); add16(&hl, de); return 0;
    case 0x1a: cyc(7);  *a = load8(de); return 0;
    case 0x1b: cyc(6);  de--; return 0;
    case 0x1c: cyc(4);  inc(e); return 0;
    case 0x1d: cyc(4);  dec(e); return 0;
    case 0x1e: cyc(7);  *e = fetch_n(); return 0;
    case 0x1f: cyc(4);  rra(); return 0;
    case 0x20: cyc( branch(!if_z) ); return 0;
    case 0x21: cyc(10); hl = fetch_nn(); return 0;
    case 0x22: cyc(16); store16(fetch_nn(), hl); return MEM_CHG;
    case 0x23: cyc(6);  hl++; return 0;
    case 0x24: cyc(4);  inc(h); return 0;
    case 0x25: cyc(4);  dec(h); return 0;
    case 0x26: cyc(7);  *h = fetch_n(); return 0;
    case 0x27: cyc(4);  daa(); return 0;
    case 0x28: cyc( branch(if_z) ); return 0;
    case 0x29: cyc(11); add16(&hl, hl); return 0;
    case 0x2a: cyc(16); hl = load16(fetch_nn()); return 0;
    case 0x2b: cyc(6);  hl--; return 0;
    case 0x2c: cyc(4);  inc(l); return 0;
    case 0x2d: cyc(4);  dec(l); return 0;
    case 0x2e: cyc(7);  *l = fetch_n(); return 0;
    case 0x2f: cyc(4);  *a = ~*a; assign_h(1); assign_n(1); return 0;
    case 0x30: cyc( branch(!if_c) ); return 0;
    case 0x31: cyc(10); sp = fetch_nn(); return 0;
    case 0x32: cyc(13); store8(fetch_nn(), *a); return MEM_CHG;
    case 0x33: cyc(6);  sp++; return 0;
    case 0x34: cyc(11); t = load8(hl); inc(&t); m_tcount++; store8(hl, t); return MEM_CHG;
    case 0x35: cyc(11); t = load8(hl); dec(&t); m_tcount++; store8(hl, t); return MEM_CHG;
    case 0x36: cyc(10); store8(hl, fetch_n()); return MEM_CHG;
    case 0x37: cyc(4);  assign_c(1); assign_h(0); assign_n(0); return 0;
    case 0x38: cyc( branch(if_c) ); return 0;
    case 0x39: cyc(11); add16(&hl, sp); return 0;
    case 0x3a: cyc(13); *a = load8(fetch_nn()); return 0;
    case 0x3b: cyc(6);  sp--; return 0;
    case 0x3c: cyc(4);  inc(a); return 0;
    case 0x3d: cyc(4);  dec(a); return 0;
    case 0x3e: cyc(7);  *a = fetch_n(); return 0;
    case 0x3f: cyc(4);  assign_h(carry); assign_c(!carry); assign_n(0); return 0;
    }
	return 0;
}

int Zi80::main01(u8 op)
{
    switch (op) {
    case 0x40: cyc(4);  return 0;
    case 0x41: cyc(4);  *b = *c; return 0;
    case 0x42: cyc(4);  *b = *d; return 0;
    case 0x43: cyc(4);  *b = *e; return 0;
    case 0x44: cyc(4);  *b = *h; return 0;
    case 0x45: cyc(4);  *b = *l; return 0;
    case 0x46: cyc(7);  *b = load8(hl); return 0;
    case 0x47: cyc(4);  *b = *a; return 0;
    case 0x48: cyc(4);  *c = *b; return 0;
    case 0x49: cyc(4);  return 0;
    case 0x4a: cyc(4);  *c = *d; return 0;
    case 0x4b: cyc(4);  *c = *e; return 0;
    case 0x4c: cyc(4);  *c = *h; return 0;
    case 0x4d: cyc(4);  *c = *l; return 0;
    case 0x4e: cyc(7);  *c = load8(hl); return 0;
    case 0x4f: cyc(4);  *c = *a; return 0;
    case 0x50: cyc(4);  *d = *b; return 0;
    case 0x51: cyc(4);  *d = *c; return 0;
    case 0x52: cyc(4);  return 0;
    case 0x53: cyc(4);  *d = *e; return 0;
    case 0x54: cyc(4);  *d = *h; return 0;
    case 0x55: cyc(4);  *d = *l; return 0;
    case 0x56: cyc(7);  *d = load8(hl); return 0;
    case 0x57: cyc(4);  *d = *a; return 0;
    case 0x58: cyc(4);  *e = *b; return 0;
    case 0x59: cyc(4);  *e = *c; return 0;
    case 0x5a: cyc(4);  *e = *d; return 0;
    case 0x5b: cyc(4);  return 0;
    case 0x5c: cyc(4);  *e = *h; return 0;
    case 0x5d: cyc(4);  *e = *l; return 0;
    case 0x5e: cyc(7);  *e = load8(hl); return 0;
    case 0x5f: cyc(4);  *e = *a; return 0;
    case 0x60: cyc(4);  *h = *b; return 0;
    case 0x61: cyc(4);  *h = *c; return 0;
    case 0x62: cyc(4);  *h = *d; return 0;
    case 0x63: cyc(4);  *h = *e; return 0;
    case 0x64: cyc(4);  return 0;
    case 0x65: cyc(4);  *h = *l; return 0;
    case 0x66: cyc(7);  *h = load8(hl); return 0;
    case 0x67: cyc(4);  *h = *a; return 0;
    case 0x68: cyc(4);  *l = *b; return 0;
    case 0x69: cyc(4);  *l = *c; return 0;
    case 0x6a: cyc(4);  *l = *d; return 0;
    case 0x6b: cyc(4);  *l = *e; return 0;
    case 0x6c: cyc(4);  *l = *h; return 0;
    case 0x6d: cyc(4);  return 0;
    case 0x6e: cyc(7);  *l = load8(hl); return 0;
    case 0x6f: cyc(4);  *l = *a; return 0;
    case 0x70: cyc(7);  store8(hl, *b); return MEM_CHG;
    case 0x71: cyc(7);  store8(hl, *c); return MEM_CHG;
    case 0x72: cyc(7);  store8(hl, *d); return MEM_CHG;
    case 0x73: cyc(7);  store8(hl, *e); return MEM_CHG;
    case 0x74: cyc(7);  store8(hl, *h); return MEM_CHG;
    case 0x75: cyc(7);  store8(hl, *l); return MEM_CHG;
    case 0x76: cyc(4);  pc--; return RHALT;
    case 0x77: cyc(7);  store8(hl, *a); return MEM_CHG;
    case 0x78: cyc(4);  *a = *b; return 0;
    case 0x79: cyc(4);  *a = *c; return 0;
    case 0x7a: cyc(4);  *a = *d; return 0;
    case 0x7b: cyc(4);  *a = *e; return 0;
    case 0x7c: cyc(4);  *a = *h; return 0;
    case 0x7d: cyc(4);  *a = *l; return 0;
    case 0x7e: cyc(7);  *a = load8(hl); return 0;
    case 0x7f: cyc(4);  return 0;
    }
	return 0;
}

int Zi80::main10(u8 op)
{
    switch (op) {
    case 0x80: cyc(4);  add(a, *b, 0, 1); return 0;
    case 0x81: cyc(4);  add(a, *c, 0, 1); return 0;
    case 0x82: cyc(4);  add(a, *d, 0, 1); return 0;
    case 0x83: cyc(4);  add(a, *e, 0, 1); return 0;
    case 0x84: cyc(4);  add(a, *h, 0, 1); return 0;
    case 0x85: cyc(4);  add(a, *l, 0, 1); return 0;
    case 0x86: cyc(7);  add(a, load8(hl), 0, 1); return 0;
    case 0x87: cyc(4);  add(a, *a, 0, 1); return 0;
    case 0x88: cyc(4);  add(a, *b, carry, 1); return 0;
    case 0x89: cyc(4);  add(a, *c, carry, 1); return 0;
    case 0x8a: cyc(4);  add(a, *d, carry, 1); return 0;
    case 0x8b: cyc(4);  add(a, *e, carry, 1); return 0;
    case 0x8c: cyc(4);  add(a, *h, carry, 1); return 0;
    case 0x8d: cyc(4);  add(a, *l, carry, 1); return 0;
    case 0x8e: cyc(7);  add(a, load8(hl), carry, 1); return 0;
    case 0x8f: cyc(4);  add(a, *a, carry, 1); return 0;
    case 0x90: cyc(4);  sub(a, *b, 0, 1); return 0;
    case 0x91: cyc(4);  sub(a, *c, 0, 1); return 0;
    case 0x92: cyc(4);  sub(a, *d, 0, 1); return 0;
    case 0x93: cyc(4);  sub(a, *e, 0, 1); return 0;
    case 0x94: cyc(4);  sub(a, *h, 0, 1); return 0;
    case 0x95: cyc(4);  sub(a, *l, 0, 1); return 0;
    case 0x96: cyc(7);  sub(a, load8(hl), 0, 1); return 0;
    case 0x97: cyc(4);  sub(a, *a, 0, 1); return 0;
    case 0x98: cyc(4);  sub(a, *b, carry, 1); return 0;
    case 0x99: cyc(4);  sub(a, *c, carry, 1); return 0;
    case 0x9a: cyc(4);  sub(a, *d, carry, 1); return 0;
    case 0x9b: cyc(4);  sub(a, *e, carry, 1); return 0;
    case 0x9c: cyc(4);  sub(a, *h, carry, 1); return 0;
    case 0x9d: cyc(4);  sub(a, *l, carry, 1); return 0;
    case 0x9e: cyc(7);  sub(a, load8(hl), carry, 1); return 0;
    case 0x9f: cyc(4);  sub(a, *a, carry, 1); return 0;
    case 0xa0: cyc(4);  andz(*b); return 0;
    case 0xa1: cyc(4);  andz(*c); return 0;
    case 0xa2: cyc(4);  andz(*d); return 0;
    case 0xa3: cyc(4);  andz(*e); return 0;
    case 0xa4: cyc(4);  andz(*h); return 0;
    case 0xa5: cyc(4);  andz(*l); return 0;
    case 0xa6: cyc(7);  andz(load8(hl)); return 0;
    case 0xa7: cyc(4);  andz(*a); return 0;
    case 0xa8: cyc(4);  xorz(*b); return 0;
    case 0xa9: cyc(4);  xorz(*c); return 0;
    case 0xaa: cyc(4);  xorz(*d); return 0;
    case 0xab: cyc(4);  xorz(*e); return 0;
    case 0xac: cyc(4);  xorz(*h); return 0;
    case 0xad: cyc(4);  xorz(*l); return 0;
    case 0xae: cyc(7);  xorz(load8(hl)); return 0;
    case 0xaf: cyc(4);  xorz(*a); return 0;
    case 0xb0: cyc(4);  orz(*b); return 0;
    case 0xb1: cyc(4);  orz(*c); return 0;
    case 0xb2: cyc(4);  orz(*d); return 0;
    case 0xb3: cyc(4);  orz(*e); return 0;
    case 0xb4: cyc(4);  orz(*h); return 0;
    case 0xb5: cyc(4);  orz(*l); return 0;
    case 0xb6: cyc(7);  orz(load8(hl)); return 0;
    case 0xb7: cyc(4);  orz(*a); return 0;
    case 0xb8: cyc(4);  cp(*b); return 0;
    case 0xb9: cyc(4);  cp(*c); return 0;
    case 0xba: cyc(4);  cp(*d); return 0;
    case 0xbb: cyc(4);  cp(*e); return 0;
    case 0xbc: cyc(4);  cp(*h); return 0;
    case 0xbd: cyc(4);  cp(*l); return 0;
    case 0xbe: cyc(7);  cp(load8(hl)); return 0;
    case 0xbf: cyc(4);  cp(*a); return 0;
    }
	return 0;
}

int Zi80::main11(u8 op)
{
    switch (op) {
    case 0xc0: cyc(5);  if (!if_z) {pop(&pc); cyc(6); } return 0;
    case 0xc1: cyc(10); pop(&bc); return 0;
    case 0xc2: cyc(10); jump(!if_z); return 0;
    case 0xc3: cyc(10); jump(1); return 0;
    case 0xc4: cyc(10); call(!if_z); return MEM_CHG;
    case 0xc5: cyc(11); push(bc); return MEM_CHG;
    case 0xc6: cyc(7);  add(a, fetch_n(), 0, 1); return 0;
    case 0xc7: cyc(11); push(pc); pc = 0; return MEM_CHG;
    case 0xc8: cyc(5);  if (if_z) {pop(&pc); cyc(6);} return 0;
    case 0xc9: cyc(10); pop(&pc); return 0;
    case 0xca: cyc(10); jump(if_z); return 0;
    case 0xcb: return cb();
    case 0xcc: cyc(10); call(if_z); return MEM_CHG;
    case 0xcd: cyc(10); call(1); return MEM_CHG;  /* call() will add the 7 tstates */
    case 0xce: cyc(7);  add(a, fetch_n(), carry, 1); return 0;
    case 0xcf: cyc(11); push(pc); pc = 8; return MEM_CHG;
    case 0xd0: cyc(5);  if (!if_c) { pop(&pc); cyc(6); } return 0;
    case 0xd1: cyc(10); pop(&de); return 0;
    case 0xd2: cyc(10); jump(!if_c); return 0;
    case 0xd3: cyc(11); out(fetch_n(), *a); return IO_CHG;
    case 0xd4: cyc(10); call(!if_c); return MEM_CHG;
    case 0xd5: cyc(11); push(de); return MEM_CHG;
    case 0xd6: cyc(7);  sub(a, fetch_n(), 0, 1); return 0;
    case 0xd7: cyc(11); push(pc); pc = 0x10; return MEM_CHG;
    case 0xd8: cyc(5);  if (if_c) { pop(&pc);  cyc(6); } return 0;
    case 0xd9: cyc(4);  ex(&bc, &bcp); ex(&de, &dep); ex(&hl, &hlp); return 0;
    case 0xda: cyc(10); jump(if_c); return 0;
    case 0xdb: cyc(11); in(fetch_n(), a, 0); return IO_CHG;
    case 0xdc: cyc(10); call(if_c); return MEM_CHG;
    case 0xdd: return ixiy(&ix);
    case 0xde: cyc(7);  sub(a, fetch_n(), carry, 1); return 0;
    case 0xdf: cyc(11); push(pc); pc = 0x18; return MEM_CHG;
    case 0xe0: cyc(5);  if (!if_pv) { pop(&pc); cyc(6); } return 0;
    case 0xe1: cyc(10); pop(&hl); return 0;
    case 0xe2: cyc(10); jump(!if_pv); return 0;
    case 0xe3: cyc(19); ex_memsp(&hl); return MEM_CHG;
    case 0xe4: cyc(10); call(!if_pv); return MEM_CHG;
    case 0xe5: cyc(11); push(hl); return MEM_CHG;
    case 0xe6: cyc(7);  andz(fetch_n()); return 0;
    case 0xe7: cyc(11); push(pc); pc = 0x20; return MEM_CHG;
    case 0xe8: cyc(5);  if (if_pv) { pop(&pc); cyc(6); } return 0;
    case 0xe9: cyc(4);  pc = hl; return 0;
    case 0xea: cyc(10); jump(if_pv); return 0;
    case 0xeb: cyc(4);  ex(&de, &hl); return 0;
    case 0xec: cyc(10); call(if_pv); return MEM_CHG;
    case 0xed: return ed();
    case 0xee: cyc(7);  xorz(fetch_n()); return 0;
    case 0xef: cyc(11); push(pc); pc = 0x28; return MEM_CHG;
    case 0xf0: cyc(5);  if (!if_s) { pop(&pc); cyc(6); } return 0;
    case 0xf1: cyc(10); pop(&af); return 0;
    case 0xf2: cyc(10); jump(!if_s); return 0;
    case 0xf3: cyc(4);  m_reg.IFF1 = m_reg.IFF2 = 0; return 0;
    case 0xf4: cyc(10); call(!if_s); return 0;
    case 0xf5: cyc(11); push(af); return MEM_CHG;
    case 0xf6: cyc(7);  orz(fetch_n()); return 0;
    case 0xf7: cyc(11); push(pc); pc = 0x30; return MEM_CHG;
    case 0xf8: cyc(5);  if (if_s) { pop(&pc); cyc(6); } return 0;
    case 0xf9: cyc(6);  sp = hl; return 0;
    case 0xfa: cyc(10); jump(if_s); return 0;
    case 0xfb: cyc(4);  m_reg.IFF1 = m_reg.IFF2 = 1; return 0; // XXX - actually, interrupts not enabled until after the instruction after EI
    case 0xfc: cyc(10); call(if_s); return MEM_CHG;
    case 0xfd: return ixiy(&iy);
    case 0xfe: cyc(7);  cp(fetch_n()); return 0;
    case 0xff: cyc(11); push(pc); pc = 0x38; return MEM_CHG;
    }
	return 0;
}

int Zi80::cb()
{
	u8 op = fetch_opcode();

    switch (op & 0xc0) {
    case 0x00: return cb00(op);
    case 0x40: return cb01(op);
    case 0x80: return cb10(op);
    case 0xc0: return cb11(op);
    }
	return 0;
}

int Zi80::cb00(u8 op)
{
    u8 t;

    switch (op) {
    case 0x00: cyc(8);  rlc(b); return 0;
    case 0x01: cyc(8);  rlc(c); return 0;
    case 0x02: cyc(8);  rlc(d); return 0;
    case 0x03: cyc(8);  rlc(e); return 0;
    case 0x04: cyc(8);  rlc(h); return 0;
    case 0x05: cyc(8);  rlc(l); return 0;
    case 0x06: cyc(15); t = load8(hl); rlc(&t); store8(hl, t); return MEM_CHG;
    case 0x07: cyc(8);  rlc(a); return 0;
    case 0x08: cyc(8);  rrc(b); return 0;
    case 0x09: cyc(8);  rrc(c); return 0;
    case 0x0a: cyc(8);  rrc(d); return 0;
    case 0x0b: cyc(8);  rrc(e); return 0;
    case 0x0c: cyc(8);  rrc(h); return 0;
    case 0x0d: cyc(8);  rrc(l); return 0;
    case 0x0e: cyc(15); t = load8(hl); rrc(&t); store8(hl, t); return MEM_CHG;
    case 0x0f: cyc(8);  rrc(a); return 0;
    case 0x10: cyc(8);  rl(b); return 0;
    case 0x11: cyc(8);  rl(c); return 0;
    case 0x12: cyc(8);  rl(d); return 0;
    case 0x13: cyc(8);  rl(e); return 0;
    case 0x14: cyc(8);  rl(h); return 0;
    case 0x15: cyc(8);  rl(l); return 0;
    case 0x16: cyc(15); t = load8(hl); rl(&t); store8(hl, t); return MEM_CHG;
    case 0x17: cyc(8);  rl(a); return 0;
    case 0x18: cyc(8);  rr(b); return 0;
    case 0x19: cyc(8);  rr(c); return 0;
    case 0x1a: cyc(8);  rr(d); return 0;
    case 0x1b: cyc(8);  rr(e); return 0;
    case 0x1c: cyc(8);  rr(h); return 0;
    case 0x1d: cyc(8);  rr(l); return 0;
    case 0x1e: cyc(15); t = load8(hl); rr(&t); store8(hl, t); return MEM_CHG;
    case 0x1f: cyc(8);  rr(a); return 0;
    case 0x20: cyc(8);  sla(b); return 0;
    case 0x21: cyc(8);  sla(c); return 0;
    case 0x22: cyc(8);  sla(d); return 0;
    case 0x23: cyc(8);  sla(e); return 0;
    case 0x24: cyc(8);  sla(h); return 0;
    case 0x25: cyc(8);  sla(l); return 0;
    case 0x26: cyc(15); t = load8(hl); sla(&t); store8(hl, t); return MEM_CHG;
    case 0x27: cyc(8);  sla(a); return 0;
    case 0x28: cyc(8);  sra(b); return 0;
    case 0x29: cyc(8);  sra(c); return 0;
    case 0x2a: cyc(8);  sra(d); return 0;
    case 0x2b: cyc(8);  sra(e); return 0;
    case 0x2c: cyc(8);  sra(h); return 0;
    case 0x2d: cyc(8);  sra(l); return 0;
    case 0x2e: cyc(15); t = load8(hl); sra(&t); store8(hl, t); return MEM_CHG;
    case 0x2f: cyc(8);  sra(a); return 0;
    case 0x38: cyc(8);  srl(b); return 0;
    case 0x39: cyc(8);  srl(c); return 0;
    case 0x3a: cyc(8);  srl(d); return 0;
    case 0x3b: cyc(8);  srl(e); return 0;
    case 0x3c: cyc(8);  srl(h); return 0;
    case 0x3d: cyc(8);  srl(l); return 0;
    case 0x3e: cyc(15); t = load8(hl); srl(&t); store8(hl, t); return MEM_CHG;
    case 0x3f: cyc(8);  srl(a); return 0;
    default: return ILL;
    }
	return 0;
}

int Zi80::cb01(u8 op)
{
    switch (op) {
    case 0x40: cyc(8);  bit(*b, 0); return 0;
    case 0x41: cyc(8);  bit(*c, 0); return 0;
    case 0x42: cyc(8);  bit(*d, 0); return 0;
    case 0x43: cyc(8);  bit(*e, 0); return 0;
    case 0x44: cyc(8);  bit(*h, 0); return 0;
    case 0x45: cyc(8);  bit(*l, 0); return 0;
    case 0x46: cyc(12); bit(load8(hl), 0); return 0;
    case 0x47: cyc(8);  bit(*a, 0); return 0;
    case 0x48: cyc(8);  bit(*b, 1); return 0;
    case 0x49: cyc(8);  bit(*c, 1); return 0;
    case 0x4a: cyc(8);  bit(*d, 1); return 0;
    case 0x4b: cyc(8);  bit(*e, 1); return 0;
    case 0x4c: cyc(8);  bit(*h, 1); return 0;
    case 0x4d: cyc(8);  bit(*l, 1); return 0;
    case 0x4e: cyc(12); bit(load8(hl), 1); return 0;
    case 0x4f: cyc(8);  bit(*a, 1); return 0;
    case 0x50: cyc(8);  bit(*b, 2); return 0;
    case 0x51: cyc(8);  bit(*c, 2); return 0;
    case 0x52: cyc(8);  bit(*d, 2); return 0;
    case 0x53: cyc(8);  bit(*e, 2); return 0;
    case 0x54: cyc(8);  bit(*h, 2); return 0;
    case 0x55: cyc(8);  bit(*l, 2); return 0;
    case 0x56: cyc(12); bit(load8(hl), 2); return 0;
    case 0x57: cyc(8);  bit(*a, 2); return 0;
    case 0x58: cyc(8);  bit(*b, 3); return 0;
    case 0x59: cyc(8);  bit(*c, 3); return 0;
    case 0x5a: cyc(8);  bit(*d, 3); return 0;
    case 0x5b: cyc(8);  bit(*e, 3); return 0;
    case 0x5c: cyc(8);  bit(*h, 3); return 0;
    case 0x5d: cyc(8);  bit(*l, 3); return 0;
    case 0x5e: cyc(12); bit(load8(hl), 3); return 0;
    case 0x5f: cyc(8);  bit(*a, 3); return 0;
    case 0x60: cyc(8);  bit(*b, 4); return 0;
    case 0x61: cyc(8);  bit(*c, 4); return 0;
    case 0x62: cyc(8);  bit(*d, 4); return 0;
    case 0x63: cyc(8);  bit(*e, 4); return 0;
    case 0x64: cyc(8);  bit(*h, 4); return 0;
    case 0x65: cyc(8);  bit(*l, 4); return 0;
    case 0x66: cyc(12); bit(load8(hl), 4); return 0;
    case 0x67: cyc(8);  bit(*a, 4); return 0;
    case 0x68: cyc(8);  bit(*b, 5); return 0;
    case 0x69: cyc(8);  bit(*c, 5); return 0;
    case 0x6a: cyc(8);  bit(*d, 5); return 0;
    case 0x6b: cyc(8);  bit(*e, 5); return 0;
    case 0x6c: cyc(8);  bit(*h, 5); return 0;
    case 0x6d: cyc(8);  bit(*l, 5); return 0;
    case 0x6e: cyc(12); bit(load8(hl), 5); return 0;
    case 0x6f: cyc(8);  bit(*a, 5); return 0;
    case 0x70: cyc(8);  bit(*b, 6); return 0;
    case 0x71: cyc(8);  bit(*c, 6); return 0;
    case 0x72: cyc(8);  bit(*d, 6); return 0;
    case 0x73: cyc(8);  bit(*e, 6); return 0;
    case 0x74: cyc(8);  bit(*h, 6); return 0;
    case 0x75: cyc(8);  bit(*l, 6); return 0;
    case 0x76: cyc(12); bit(load8(hl), 6); return 0;
    case 0x77: cyc(8);  bit(*a, 6); return 0;
    case 0x78: cyc(8);  bit(*b, 7); return 0;
    case 0x79: cyc(8);  bit(*c, 7); return 0;
    case 0x7a: cyc(8);  bit(*d, 7); return 0;
    case 0x7b: cyc(8);  bit(*e, 7); return 0;
    case 0x7c: cyc(8);  bit(*h, 7); return 0;
    case 0x7d: cyc(8);  bit(*l, 7); return 0;
    case 0x7e: cyc(12); bit(load8(hl), 7); return 0;
    case 0x7f: cyc(8);  bit(*a, 7); return 0;
    }
	return 0;
}

int Zi80::cb10(u8 op)
{
    u8 t;

    switch (op) {
    case 0x80: cyc(8);  res(b, 0); return 0;
    case 0x81: cyc(8);  res(c, 0); return 0;
    case 0x82: cyc(8);  res(d, 0); return 0;
    case 0x83: cyc(8);  res(e, 0); return 0;
    case 0x84: cyc(8);  res(h, 0); return 0;
    case 0x85: cyc(8);  res(l, 0); return 0;
    case 0x86: cyc(15); t = load8(hl); res(&t, 0); store8(hl, t); return MEM_CHG;
    case 0x87: cyc(8);  res(a, 0); return 0;
    case 0x88: cyc(8);  res(b, 1); return 0;
    case 0x89: cyc(8);  res(c, 1); return 0;
    case 0x8a: cyc(8);  res(d, 1); return 0;
    case 0x8b: cyc(8);  res(e, 1); return 0;
    case 0x8c: cyc(8);  res(h, 1); return 0;
    case 0x8d: cyc(8);  res(l, 1); return 0;
    case 0x8e: cyc(15); t = load8(hl); res(&t, 1); store8(hl, t); return MEM_CHG;
    case 0x8f: cyc(8);  res(a, 1); return 0;
    case 0x90: cyc(8);  res(b, 2); return 0;
    case 0x91: cyc(8);  res(c, 2); return 0;
    case 0x92: cyc(8);  res(d, 2); return 0;
    case 0x93: cyc(8);  res(e, 2); return 0;
    case 0x94: cyc(8);  res(h, 2); return 0;
    case 0x95: cyc(8);  res(l, 2); return 0;
    case 0x96: cyc(15); t = load8(hl); res(&t, 2); store8(hl, t); return MEM_CHG;
    case 0x97: cyc(8);  res(a, 2); return 0;
    case 0x98: cyc(8);  res(b, 3); return 0;
    case 0x99: cyc(8);  res(c, 3); return 0;
    case 0x9a: cyc(8);  res(d, 3); return 0;
    case 0x9b: cyc(8);  res(e, 3); return 0;
    case 0x9c: cyc(8);  res(h, 3); return 0;
    case 0x9d: cyc(8);  res(l, 3); return 0;
    case 0x9e: cyc(15); t = load8(hl); res(&t, 3); store8(hl, t); return MEM_CHG;
    case 0x9f: cyc(8);  res(a, 3); return 0;
    case 0xa0: cyc(8);  res(b, 4); return 0;
    case 0xa1: cyc(8);  res(c, 4); return 0;
    case 0xa2: cyc(8);  res(d, 4); return 0;
    case 0xa3: cyc(8);  res(e, 4); return 0;
    case 0xa4: cyc(8);  res(h, 4); return 0;
    case 0xa5: cyc(8);  res(l, 4); return 0;
    case 0xa6: cyc(15); t = load8(hl); res(&t, 4); store8(hl, t); return MEM_CHG;
    case 0xa7: cyc(8);  res(a, 4); return 0;
    case 0xa8: cyc(8);  res(b, 5); return 0;
    case 0xa9: cyc(8);  res(c, 5); return 0;
    case 0xaa: cyc(8);  res(d, 5); return 0;
    case 0xab: cyc(8);  res(e, 5); return 0;
    case 0xac: cyc(8);  res(h, 5); return 0;
    case 0xad: cyc(8);  res(l, 5); return 0;
    case 0xae: cyc(15); t = load8(hl); res(&t, 5); store8(hl, t); return MEM_CHG;
    case 0xaf: cyc(8);  res(a, 5); return 0;
    case 0xb0: cyc(8);  res(b, 6); return 0;
    case 0xb1: cyc(8);  res(c, 6); return 0;
    case 0xb2: cyc(8);  res(d, 6); return 0;
    case 0xb3: cyc(8);  res(e, 6); return 0;
    case 0xb4: cyc(8);  res(h, 6); return 0;
    case 0xb5: cyc(8);  res(l, 6); return 0;
    case 0xb6: cyc(15); t = load8(hl); res(&t, 6); store8(hl, t); return MEM_CHG;
    case 0xb7: cyc(8);  res(a, 6); return 0;
    case 0xb8: cyc(8);  res(b, 7); return 0;
    case 0xb9: cyc(8);  res(c, 7); return 0;
    case 0xba: cyc(8);  res(d, 7); return 0;
    case 0xbb: cyc(8);  res(e, 7); return 0;
    case 0xbc: cyc(8);  res(h, 7); return 0;
    case 0xbd: cyc(8);  res(l, 7); return 0;
    case 0xbe: cyc(15); t = load8(hl); res(&t, 7); store8(hl, t); return MEM_CHG;
    case 0xbf: cyc(8);  res(a, 7); return 0;
    }
	return 0;
}

int Zi80::cb11(u8 op)
{
    u8 t;

    switch (op) {
    case 0xc0: cyc(8);  set(b, 0); return 0;
    case 0xc1: cyc(8);  set(c, 0); return 0;
    case 0xc2: cyc(8);  set(d, 0); return 0;
    case 0xc3: cyc(8);  set(e, 0); return 0;
    case 0xc4: cyc(8);  set(h, 0); return 0;
    case 0xc5: cyc(8);  set(l, 0); return 0;
    case 0xc6: cyc(15); t = load8(hl); set(&t, 0); store8(hl, t); return MEM_CHG;
    case 0xc7: cyc(8);  set(a, 0); return 0;
    case 0xc8: cyc(8);  set(b, 1); return 0;
    case 0xc9: cyc(8);  set(c, 1); return 0;
    case 0xca: cyc(8);  set(d, 1); return 0;
    case 0xcb: cyc(8);  set(e, 1); return 0;
    case 0xcc: cyc(8);  set(h, 1); return 0;
    case 0xcd: cyc(8);  set(l, 1); return 0;
    case 0xce: cyc(15); t = load8(hl); set(&t, 1); store8(hl, t); return MEM_CHG;
    case 0xcf: cyc(8);  set(a, 1); return 0;
    case 0xd0: cyc(8);  set(b, 2); return 0;
    case 0xd1: cyc(8);  set(c, 2); return 0;
    case 0xd2: cyc(8);  set(d, 2); return 0;
    case 0xd3: cyc(8);  set(e, 2); return 0;
    case 0xd4: cyc(8);  set(h, 2); return 0;
    case 0xd5: cyc(8);  set(l, 2); return 0;
    case 0xd6: cyc(15);  t = load8(hl); set(&t, 2); store8(hl, t); return MEM_CHG;
    case 0xd7: cyc(8);  set(a, 2); return 0;
    case 0xd8: cyc(8);  set(b, 3); return 0;
    case 0xd9: cyc(8);  set(c, 3); return 0;
    case 0xda: cyc(8);  set(d, 3); return 0;
    case 0xdb: cyc(8);  set(e, 3); return 0;
    case 0xdc: cyc(8);  set(h, 3); return 0;
    case 0xdd: cyc(8);  set(l, 3); return 0;
    case 0xde: cyc(15); t = load8(hl); set(&t, 3); store8(hl, t); return MEM_CHG;
    case 0xdf: cyc(8);  set(a, 3); return 0;
    case 0xe0: cyc(8);  set(b, 4); return 0;
    case 0xe1: cyc(8);  set(c, 4); return 0;
    case 0xe2: cyc(8);  set(d, 4); return 0;
    case 0xe3: cyc(8);  set(e, 4); return 0;
    case 0xe4: cyc(8);  set(h, 4); return 0;
    case 0xe5: cyc(8);  set(l, 4); return 0;
    case 0xe6: cyc(15); t = load8(hl); set(&t, 4); store8(hl, t); return MEM_CHG;
    case 0xe7: cyc(8);  set(a, 4); return 0;
    case 0xe8: cyc(8);  set(b, 5); return 0;
    case 0xe9: cyc(8);  set(c, 5); return 0;
    case 0xea: cyc(8);  set(d, 5); return 0;
    case 0xeb: cyc(8);  set(e, 5); return 0;
    case 0xec: cyc(8);  set(h, 5); return 0;
    case 0xed: cyc(8);  set(l, 5); return 0;
    case 0xee: cyc(15); t = load8(hl); set(&t, 5); store8(hl, t); return MEM_CHG;
    case 0xef: cyc(8);  set(a, 5); return 0;
    case 0xf0: cyc(8);  set(b, 6); return 0;
    case 0xf1: cyc(8);  set(c, 6); return 0;
    case 0xf2: cyc(8);  set(d, 6); return 0;
    case 0xf3: cyc(8);  set(e, 6); return 0;
    case 0xf4: cyc(8);  set(h, 6); return 0;
    case 0xf5: cyc(8);  set(l, 6); return 0;
    case 0xf6: cyc(15); t = load8(hl); set(&t, 6); store8(hl, t); return MEM_CHG;
    case 0xf7: cyc(8);  set(a, 6); return 0;
    case 0xf8: cyc(8);  set(b, 7); return 0;
    case 0xf9: cyc(8);  set(c, 7); return 0;
    case 0xfa: cyc(8);  set(d, 7); return 0;
    case 0xfb: cyc(8);  set(e, 7); return 0;
    case 0xfc: cyc(8);  set(h, 7); return 0;
    case 0xfd: cyc(8);  set(l, 7); return 0;
    case 0xfe: cyc(15); t = load8(hl); set(&t, 7); store8(hl, t); return MEM_CHG;
    case 0xff: cyc(8);  set(a, 7); return 0;
    }
	return 0;
}

int Zi80::ed()
{
    u8 t;

    switch (fetch_opcode()) {
    case 0x40: cyc(12); in(*c, b, 1); return IO_CHG;
    case 0x41: cyc(12); out(*c, *b); return IO_CHG;
    case 0x42: cyc(15); sbc16(&hl, bc); return 0;
    case 0x43: cyc(20); store16(fetch_nn(), bc); return MEM_CHG;
    case 0x44: cyc(8);  t = *a; *a = 0; sub(a, t, 0, 1); return 0;
    case 0x45: cyc(14); m_reg.IFF1 = m_reg.IFF2; pop(&pc); return 0;
    case 0x46: cyc(8);  m_reg.IM = 0; return 0;
    case 0x47: cyc(9);  m_reg.I = *a; return 0;
    case 0x48: cyc(12); in(*c, c, 1); return IO_CHG;
    case 0x49: cyc(12); out(*c, *c); return IO_CHG;
    case 0x4a: cyc(15); adc16(&hl, bc); return 0;
    case 0x4b: cyc(20); bc = load16(fetch_nn()); return 0;
    case 0x4d: cyc(14); pop(&pc); return 0;  /* int acknowledge? */
    case 0x4f: cyc(9);  m_reg.R = *a; return 0;
    case 0x50: cyc(12); in(*c, d, 1); return IO_CHG;
    case 0x51: cyc(12); out(*c, *d); return IO_CHG;
    case 0x52: cyc(15); sbc16(&hl, de); return 0;
    case 0x53: cyc(20); store16(fetch_nn(), de); return MEM_CHG;
    case 0x56: cyc(8);  m_reg.IM = 1; return 0;
    case 0x57: cyc(9); 
		*a = m_reg.I;
		assign_h(0);
		assign_n(0);
		assign_s(m_reg.I & 0x80);
		assign_z(!m_reg.I);
		assign_pv(m_reg.IFF2);
		return 0;
    case 0x58: cyc(12); in(*c, e, 1); return IO_CHG;
    case 0x59: cyc(12); out(*c, *e); return IO_CHG;
    case 0x5a: cyc(15); adc16(&hl, de); return 0;
    case 0x5b: cyc(20); de = load16(fetch_nn()); return 0;
    case 0x5e: cyc(8);  m_reg.IM = 2; return 0;
    case 0x5f:
		m_reg.R = (((m_rcounter * 0xa5) >> 8) & 0x7f); // ( m_rcounter & 0x7f );
		*a = m_reg.R;
		assign_h(0);
		assign_n(0);
		assign_s(m_reg.R & 0x80);
		assign_z(!m_reg.R);
		assign_pv(m_reg.IFF2);
		cyc(9);
		return 0;
    case 0x60: cyc(12); in(*c, h, 1); return IO_CHG;
    case 0x61: cyc(12); out(*c, *h); return IO_CHG;
    case 0x62: cyc(15); sbc16(&hl, hl); return 0;
    case 0x63: cyc(20); store16(fetch_nn(), hl); return MEM_CHG;
    case 0x67: cyc(18); 
		t = load8(hl);
		store8(hl,(u8)(((t >> 4) & 15) | (*a << 4)) );
		*a = (*a & 0xf0) | (t & 15);
		assign_s(*a & 0x80);
		assign_z(!*a);
		assign_h(0);
		assign_pv(parity(*a));
		assign_n(0);
		return MEM_CHG;
    case 0x68: cyc(12); in(*c, l, 1); return IO_CHG;
    case 0x69: cyc(12); out(*c, *l); return IO_CHG;
    case 0x6a: cyc(15); adc16(&hl, hl); return 0;
    case 0x6b: cyc(20); hl = load16(fetch_nn()); return 0;
    case 0x6f: cyc(18); 
		t = load8(hl);
		store8(hl, (u8)((t << 4) | (*a & 15)) );
		*a = (*a & 0xf0) | ((t >> 4) & 15);
		assign_s(*a & 0x80);
		assign_z(!*a);
		assign_h(0);
		assign_pv(parity(*a));
		assign_n(0);
		return MEM_CHG;
    case 0x72: cyc(15); sbc16(&hl, sp); return 0;
    case 0x73: cyc(20); store16(fetch_nn(), sp); return MEM_CHG;
    case 0x78: cyc(12); in(*c, a, 1); return IO_CHG;
    case 0x79: cyc(12); out(*c, *a); return IO_CHG;
    case 0x7a: cyc(15); adc16(&hl, sp); return 0;
    case 0x7b: cyc(20); sp = load16(fetch_nn()); return 0;
    case 0xa0: cyc(16); 
		store8(de, load8(hl++));
		de++;
		bc--;
		assign_h(0);
		assign_n(0);
		assign_pv(bc);
		return MEM_CHG;
    case 0xa1: cyc(16); 
		t = carry;
		cp(load8(hl++));
		bc--;
		assign_pv(bc);
		assign_c(t);
		return 0;
    case 0xa2: cyc(16); 
		t = load8(hl);
		in(*c, &t, 0);
		store8(hl, t);
		hl++;
		(*b)--;
		assign_z(!(*b));
		assign_n(1);
		return IO_CHG | MEM_CHG;
    case 0xa3: cyc(16); 
		out(*c, load8(hl++));
		(*b)--;
		assign_z(!(*b));
		assign_n(1);
		return IO_CHG | MEM_CHG;
    case 0xa8: cyc(16); 
		store8(de, load8(hl--));
		de--;
		bc--;
		assign_h(0);
		assign_n(0);
		assign_pv(bc);
		return MEM_CHG;
    case 0xa9: cyc(16); 
		t = carry;
		cp(load8(hl--));
		bc--;
		assign_pv(bc);
		assign_c(t);
		return 0;
    case 0xaa: cyc(16); 
		t = load8(hl);
		in(*c, &t, 0);
		store8(hl, t);
		hl--;
		(*b)--;
		assign_z(!(*b));
		assign_n(1);
		return IO_CHG | MEM_CHG;
    case 0xab: cyc(16); 
		out(*c, load8(hl--));
		(*b)--;
		assign_z(!(*b));
		assign_n(1);
		return IO_CHG | MEM_CHG;
    case 0xb0: /* LDIR */
		store8(de, load8(hl++));
		de++;
		bc--;
		assign_h(0);
		assign_n(0);
		assign_pv(bc);
		if ( bc == 0 ) { cyc( 16 ); return MEM_CHG; }
		cyc(21);
		pc -= 2;
		return MEM_CHG;
    case 0xb1: /* CPIR */
		t = carry;
	    cp(load8(hl++));
	    bc--;
	    assign_pv(bc);
	    assign_c(t);
		if ( bc == 0 || if_z ) { cyc( 16 ); return 0; }
		cyc(21);
		pc -= 2;
		return 0;
    case 0xb2: cyc(12); 
		// XXX - looping way not accurate.  Highly broken in cycles, etc.
		do {
		    in(*c, &t, 0);
			store8(hl++, t);
		    (*b)--;
		    assign_z(!(*b));
		    assign_n(1);
		} while (*b);
		return IO_CHG | MEM_CHG;
    case 0xb3: cyc(12); 
		// XXX - looping way not accurate.  Highly broken in cycles, etc.
		do {
		    out(*c, load8(hl++));
		    (*b)--;
		    assign_z(!(*b));
		    assign_n(1);
		} while (*b);
		return IO_CHG | MEM_CHG;
    case 0xb8:	/* LDDR */
		store8(de, load8(hl--));
		de--;
		bc--;
		assign_h(0);
		assign_n(0);
		assign_pv(bc);
		if ( bc == 0 ) { cyc( 16 ); return MEM_CHG; }
		cyc(21);
		pc -= 2;
		return MEM_CHG;
    case 0xb9: /* CPDR */ 
		t = carry;
	    cp(load8(hl--));
	    bc--;
	    assign_pv(bc);
	    assign_c(t);
		if ( bc == 0 || if_z ) { cyc( 16 ); return 0; }
		pc -= 2;
		return 0;
    case 0xba: cyc(12); 
		// XXX - looping way not accurate.  Highly broken in cycles, etc.
		do {
		    in(*c, &t, 0);
			store8(hl--, t);
		    (*b)--;
		    assign_z(!(*b));
		    assign_n(1);
		} while (*b);
		return IO_CHG | MEM_CHG;
    case 0xbb: cyc(12); 
		// XXX - looping way not accurate.  Highly broken in cycles, etc.
		do {
		    out(*c, load8(hl));
		    (*b)--;
		    assign_z(!(*b));
		    assign_n(1);
		} while (*b);
		return IO_CHG | MEM_CHG;
    default: return ILL;
    }
	return 0;
}

int Zi80::ixiy(u16 *xy)
{
    u16 ixy_d;
    u8 t;

    switch (fetch_opcode()) {
    case 0x09: cyc(15); add16(xy, bc); return 0;
    case 0x19: cyc(15); add16(xy, de); return 0;
    case 0x21: cyc(14); *xy = fetch_nn(); return 0;
    case 0x22: cyc(20); store16(fetch_nn(), *xy); return MEM_CHG;
    case 0x23: cyc(10); (*xy)++; return 0;
    case 0x29: cyc(15); add16(xy, *xy); return 0;
    case 0x2a: cyc(20); *xy = load16(fetch_nn()); return 0;
    case 0x2b: cyc(10); (*xy)--; return 0;
    case 0x34: cyc(23); t = load8(ixy_d = ixy(xy)); inc(&t); store8(ixy_d, t); return MEM_CHG;
    case 0x35: cyc(23); t = load8(ixy_d = ixy(xy)); dec(&t); store8(ixy_d, t); return MEM_CHG;
		/* following is a bit off; fetch_n() likely happens a cycle earlier. */
    case 0x36: cyc(19); ixy_d = ixy(xy); m_tcount -= 4; store8(ixy_d, fetch_n()); return MEM_CHG;
    case 0x39: cyc(15); add16(xy, sp); return 0;
    case 0x46: cyc(19); *b = load8(ixy(xy)); return 0;
    case 0x4e: cyc(19); *c = load8(ixy(xy)); return 0;
    case 0x56: cyc(19); *d = load8(ixy(xy)); return 0;
    case 0x5e: cyc(19); *e = load8(ixy(xy)); return 0;
    case 0x66: cyc(19); *h = load8(ixy(xy)); return 0;
    case 0x6e: cyc(19); *l = load8(ixy(xy)); return 0;
    case 0x70: cyc(19); store8(ixy(xy), *b); return MEM_CHG;
    case 0x71: cyc(19); store8(ixy(xy), *c); return MEM_CHG;
    case 0x72: cyc(19); store8(ixy(xy), *d); return MEM_CHG;
    case 0x73: cyc(19); store8(ixy(xy), *e); return MEM_CHG;
    case 0x74: cyc(19); store8(ixy(xy), *h); return MEM_CHG;
    case 0x75: cyc(19); store8(ixy(xy), *l); return MEM_CHG;
    case 0x77: cyc(19); store8(ixy(xy), *a); return MEM_CHG;
    case 0x7e: cyc(19); *a = load8(ixy(xy)); return 0;
    case 0x86: cyc(19); add(a, load8(ixy(xy)), 0, 1); return 0;
    case 0x8e: cyc(19); add(a, load8(ixy(xy)), carry, 1); return 0;
    case 0x96: cyc(19); sub(a, load8(ixy(xy)), 0, 1); return 0;
    case 0x9e: cyc(19); sub(a, load8(ixy(xy)), carry, 1); return 0;
    case 0xa6: cyc(19); andz(load8(ixy(xy))); return 0;
    case 0xae: cyc(19); xorz(load8(ixy(xy))); return 0;
    case 0xb6: cyc(19); orz(load8(ixy(xy))); return 0;
    case 0xbe: cyc(19); cp(load8(ixy(xy))); return 0;
    case 0xe1: cyc(14); pop(xy); return 0;
    case 0xe3: cyc(23); ex_memsp(xy); return MEM_CHG;
    case 0xe5: cyc(15); push(*xy); return MEM_CHG;
    case 0xe9: cyc(8);  pc = *xy; return 0;
    case 0xf9: cyc(10); sp = *xy; return 0;
    default: return ILL;
    case 0xcb:
		ixy_d = ixy(xy);
		t = load8(ixy_d);
		// Those 23 below are probably too many due to load8 above...
		switch (fetch_opcode()) {
		case 0x06: cyc(23);  rlc(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x0e: cyc(23);  rrc(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x16: cyc(23);  rl(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x1e: cyc(23);  rr(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x26: cyc(23);  sla(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x2e: cyc(23);  sra(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x3e: cyc(23);  srl(&t); store8(ixy_d, t); return MEM_CHG;
		case 0x46: cyc(20);  bit(t, 0); return 0;
		case 0x4e: cyc(20);  bit(t, 1); return 0;
		case 0x56: cyc(20);  bit(t, 2); return 0;
		case 0x5e: cyc(20);  bit(t, 3); return 0;
		case 0x66: cyc(20);  bit(t, 4); return 0;
		case 0x6e: cyc(20);  bit(t, 5); return 0;
		case 0x76: cyc(20);  bit(t, 6); return 0;
		case 0x7e: cyc(20);  bit(t, 7); return 0;
		case 0x86: cyc(23);  res(&t, 0); store8(ixy_d, t); return MEM_CHG;
		case 0x8e: cyc(23);  res(&t, 1); store8(ixy_d, t); return MEM_CHG;
		case 0x96: cyc(23);  res(&t, 2); store8(ixy_d, t); return MEM_CHG;
		case 0x9e: cyc(23);  res(&t, 3); store8(ixy_d, t); return MEM_CHG;
		case 0xa6: cyc(23);  res(&t, 4); store8(ixy_d, t); return MEM_CHG;
		case 0xae: cyc(23);  res(&t, 5); store8(ixy_d, t); return MEM_CHG;
		case 0xb6: cyc(23);  res(&t, 6); store8(ixy_d, t); return MEM_CHG;
		case 0xbe: cyc(23);  res(&t, 7); store8(ixy_d, t); return MEM_CHG;
		case 0xc6: cyc(23);  set(&t, 0); store8(ixy_d, t); return MEM_CHG;
		case 0xce: cyc(23);  set(&t, 1); store8(ixy_d, t); return MEM_CHG;
		case 0xd6: cyc(23);  set(&t, 2); store8(ixy_d, t); return MEM_CHG;
		case 0xde: cyc(23);  set(&t, 3); store8(ixy_d, t); return MEM_CHG;
		case 0xe6: cyc(23);  set(&t, 4); store8(ixy_d, t); return MEM_CHG;
		case 0xee: cyc(23);  set(&t, 5); store8(ixy_d, t); return MEM_CHG;
		case 0xf6: cyc(23);  set(&t, 6); store8(ixy_d, t); return MEM_CHG;
		case 0xfe: cyc(23);  set(&t, 7); store8(ixy_d, t); return MEM_CHG;
		default: return ILL;
		}
    }
	return 0;
}
