//
// RunLoad.cpp - detect and run loading code
//
// TRS-80 programs often had one or two bits of code at the start of a program.
// One would relocate the program to the correct loading address.  This was typically
// added by utility programs (e.g., LMOFFSET) that copied programs from tape to
// disk when the tape version would have loaded over top of the disk operating system.
// The other were specialized cassette loaders that did the job of loading the
// bulk of the program into memory.
//
// This code does a pretty decent job of detecting and running such code thus
// allowing disassemblers and the like to work with the resulting program rather
// than only showing you the loader/relocating code properly.
//
//
// Many possible sequences exist starting from loaded data.
//
// Loaded data is program, execution begins normally.
// Copy may occur.
// Then normal execution.
// Or that may be a cassette loader.
// After loading, execution may begin normally.
// Or another copy may occur.
// Then execution begins normally.
//
// In pretty much all the cases checksum error checking is done and if any
// occurs a 'C' may be written to the screen with a jump to READY or infinite loop.
//
// And many, many .cmd files are LMOFFSET versions of cassette programs that
// copy the program to where it must live and jumps to it.
//
// Some concrete examples:
//
// Time Trek Level 1
//		Loads small program that calls ROM to load game and jumps to it.
//		The loader auto-runs in typical Level 1 fashion (overwriting stack).
//
// Time Trek Level 2
//		Loads small program which uses 401e display vector to auto-run.
//		Copies it to lower memory
//		Runs custom cassette loader
//		Jumps to loaded program.
//		Note: loader does some pushes and pops thus there are some spurious bytes
//			that look like loaded data.
//
// Defense Command (and Stella Escort) from Games Pack 2
//		Loads small loader program (and a splash screen to display memory)
//		As custom loader runs updates time until finished load.
//		Moves program slighlty up in memory (using lddr)
//		Jumps into moved copy.
//
// Zaxxon
//		Very similar loader to Games Pack 2 with on-screen display.
//		And the overlapping slightly moved copy after loading.
//
// Space Warp Level 2
//		Custom loader runs.
//		Copies loaded data and zeros original (using LDI).
//		Jumps to moved copy.
//
// Space Warp Level 1
//		Small program loads; copies part of itself down low.
//		The copying part does not get copied.
//		Calls ROM to load game and jumps to it.
//
// Microchess Level 2
//		Custom loader first loads game, with single byte checksum.
//		And after a fixed number of bytes loads to screen with instructions.
//		Then jumps to game.
//
// Microchess Level 1
//		Traditional Level 1 load, but game uses low-level Level 1 cassette routines
//		to get in sync and load the next file onto the screen for the insructions.
//		Incidentally, it checks to see if it loaded properly by jumping to WHAT?
//		if NZ at program start.  Thus a loader should set Z before running Level 1
//		code.
//
// galinv1l.cmd, galinv1o.cmd, attack.3bn
//		After relocation self-modifying code makes it looks like they're
//		jumping to relocated data.  The relocated data jump compenstates by
//		considering anything less than 8 bytes in size to be self-modifying code.
//
// galinvp1.cmd, galinvp4.cmd, seadrgn3.cmd, attack.3bn
//		Partial relocation of data.  First 3 above only relocate the portion
//		that overlaps DOS.  Attack.3bn relocates the intro scene code that
//		it doesn't re-use.
//
// However, these last 4 are broken for two reasons.  First, a slight
// modification of the input data breaks detection of the relocation.
// 2nd, after conversion they may not work as .cas because they are
// loaded too low.

#include <stdio.h>
#include <string.h>

#include "RunLoad.h"

RunLoad::RunLoad()
{
	m_stopLoad = false;
	m_verbose = false;
	memset(m_write, 0, sizeof(m_write));
}

// Ultimately we'll want to run through the ROM code itself for those
// with the means.  But that's onerous and has to make assumptions that
// mean one can't conveniently load a SYSTEM tape into a Level 1 machine.
// So we just go with heuristics.

// Stop condition: 100 instructions without a write to a new location.
//
// And need (in SuperLoad) to make sure PC, HL get passed up if changed.
// And conditions to only do the first N stages and/or refuse cassette
// or relocates.  'Cept for the emulator you want to refuse relocates
// but still do them if they hide a cassette load.  Humph.

bool RunLoad::Analyze()
{
	memcpy(m_initial_load, m_write, sizeof(m_write));

	memset(m_write, 0, sizeof(m_write));
	m_clearWriteOnWrite = false;

	int start_sp = m_reg.SP;

	ClearPortChanges();

	// Remember, if we do no relocating and the load doesn't set a PC
	// we must call down with -1 in LoadMemory()

	for (m_newWriteCountdown = 100; m_newWriteCountdown > 0; m_newWriteCountdown--)
	{
		Step();

		if (m_write[m_reg.PC])
		{
			if (m_verbose)
			{
				printf("Executing written data at $%04x\n", m_reg.PC);
			}

			// Avoid being fooled by self-modifying code.  attack.3bn,
			// galinv1l.cmd, galinv1o.cmd run in to this condition.
			int written_size;
			for (written_size = 0; written_size < 8; written_size++)
			{
				if (!m_write[(m_reg.PC + written_size) & 0xffff])
					break;
			}

			if (written_size < 8)
			{
				if (m_verbose)
				{
					printf("\tIgnoring, write area only %d bytes log. Probably self-modifying code.\n",
						written_size);
				}
			}
			else
			{
				WarnPortChanges();

				CheckForPartialCopy(m_reg.PC);

				if (m_reg.SP != start_sp)
				{
					printf("stack pointer changed from $%04x to $%04x\n",
						start_sp, m_reg.SP);

					start_sp = m_reg.SP;
				}

				LoadMemory(m_mem, m_write, m_reg.PC);
				if (m_stopLoad)
				{
					return true;
				}

				memset(m_write, 0, sizeof(m_write));
				m_clearWriteOnWrite = false;

				ClearPortChanges();
			}
		}

		if (m_reg.PC < 0x4000)
		{
			// We assume CALL, but if was a jump, well, hmm.
			//printf("Call ROM $%04x\n", m_reg.PC);

			switch (m_reg.PC)
			{
			case 0x0212: // Level 2 select cassette drive
				break;
			case 0x0296: // Level 2 start tape drive and read header
				if (!ReadCassetteHeader())
				{
					return false;
				}
				break;
			case 0x022c: // Level 2 blink asterix
				break;
			case 0x0235: // Level 2 cassette read byte
				{
					unsigned char data;
					if (!ReadCassetteByte(&data))
					{
						return false;
					}
					m_reg.AF = (data << 8) | (m_reg.AF & 0xff);
				}
				break;
			case 0x01f8: // Level 2 turn off cassette drive
				// If data starts moving again then we'll respond properly
				// to the copy.
				m_clearWriteOnWrite = true;
				break;
			case 0x0ef4: // Level 1 cassette loader.  Returns Z=1 if loaded OK.
				// Only what we load is relevant.
				memset(m_write, 0, sizeof(m_write));
				if (!LoadNext(false, false))
				{
					return false;
				}
				m_reg.AF |= 0x40;
				if (m_stopLoad)
				{
					return true;
				}
				break;
			default:
				return true;
			}

			// Undo the call.
			m_mem[m_reg.PC] = 0xc9;
			Step();
		}
	}

	return true;
}

u8 RunLoad::OpcodeFetch(u16 address)
{
	return DataRead(address);
}

u8 RunLoad::IntrFetch()
{
	return 0xff;
}

u8 RunLoad::DataRead(u16 address)
{
	return m_mem[address];
}

void RunLoad::DataWrite(u16 address, u8 data)
{
	m_mem[address] = data;

	if (m_clearWriteOnWrite)
	{
		memset(m_write, 0, sizeof(m_write));
		m_clearWriteOnWrite = false;
	}

	if (!m_write[address])
	{
		m_write[address] = 1;
		m_newWriteCountdown = 100;
	}
}

u8 RunLoad::PortRead(u16 address)
{
	return 0xff;
}

void RunLoad::PortWrite(u16 address, u8 data)
{
	m_port[address & 0xff] = data;
	m_portwritten[address & 0xff] = 1;
}

void RunLoad::SetStopLoad(bool stop)
{
	m_stopLoad = stop;
}

// Given an address that has been written to, find how large a block
// of contiguous writes it lives within.  Modifies addr to be the address
// of the block's start.

int RunLoad::WrittenSize(int *addr)
{
	int low = *addr;

	for (; low >= 0 && m_write[low]; low--)
	{
	}
	low++;

	int high = *addr;

	for (; high < 65536 && m_write[high]; high++)
	{
	}

	*addr = low;

	return high - low;
}

// If the amount written is considerably less than the size of the loaded
// data we'll consider it a partial relocation and combine the written data
// with the loaded data.  But we'll drop the copied data if we can locate it.

void RunLoad::CheckForPartialCopy(int addr)
{ 
	int i;

	// Count bytes initially loaded
	int load_size = 0;
	for (i = 0; i < sizeof(m_initial_load); i++)
	{
		if (m_initial_load[i])
			load_size++;
	}

	// Figure out size of copied data assuming it is a contiguous block.

	int write_size = WrittenSize(&addr);

	if (write_size >= 0.9 * load_size)
		return;

	if (m_verbose)
		printf("partial relocate to [%04x - %04x] ", addr, addr + write_size - 1);

	// Can we find it?

	bool located = false;
	for (i = 0; i < sizeof(m_initial_load) - write_size; i++)
	{
		if (!m_initial_load[i])
			continue;

		if (i >= addr && i < addr + write_size)
			continue;

		if (memcmp(m_mem + i, m_mem + addr, write_size) == 0)
		{
			// Don't keep the source data around (unless it was possibly changed)
			for (int j = 0; j < write_size; j++)
			{
				if (!m_write[i + j])
					m_initial_load[i + j] = 0;
			}

			if (m_verbose)
				printf("from %04x\n", i);

			located = true;
		}
	}

	if (m_verbose && !located)
		printf("(source unknown)\n");

	for (i = 0; i < sizeof(m_write); i++)
	{
		if (m_initial_load[i])
			m_write[i] = 1;
	}
}

void RunLoad::ClearPortChanges()
{
	memset(m_portwritten, 0, sizeof(m_portwritten));
}

void RunLoad::WarnPortChanges()
{
	for (int i = 0; i < 256; i++)
	{
		if (m_portwritten[i])
		{
			printf("warning: port $%02x initialized to $%02x\n", i, m_port[i]);
		}
	}
}
