//
// trld -- translate/load TRS-80 binaries
//
// General use is to translate format between .cas, .cmd, .hex, etc.  Can load
// more than one binary at once to concatenate them.
//
// Initially does not preserve the original structure and doesn't handle binary
// basic files in a sensible form.
//
// Suggestions: Control over BASIC load addresses.  
// Option to leave name out of .cmd output.
//
// Future possibilities include translations that preserve as much orignial
// structure as possible.  Simple dumps.  Module offsets.
//
// Look for '%d' formatting in output file name and copy out all the programs
// found in an input file.
//

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if defined(__APPLE__) || defined(__linux__)
#define stricmp strcasecmp
#endif

#include "../../../lib/loader.h"
#include "../../../lib/SuperLoad.h"

class Loader : public SuperLoad
{
public:
	virtual void Warning(const char *message)
	{
		printf("trld: warning: '%s': %s\n", m_filename, message);
	}

	virtual void LoadMemory(unsigned char *mem, unsigned char *loaded, int pc)
	{
		SuperLoad::LoadMemory(mem, loaded, pc);
		m_stage++;
		SetStopLoad(m_stage == m_stopStage);
	}

	char *m_filename;
	int m_stage;
	int m_stopStage;
};

static void usage();

int main(int argc, char *argv[])
{
	Loader sl;
	int dump = 0;
	int pad = 1;
	int use_name = 0;

	TRS80Loader::Media outputMedia = TRS80Loader::cassette1500;
	TRS80Loader::Physical outputPhysical = TRS80Loader::justBits;
	int stopStage = 1; // i.e., don't try to run Z-80 code looking for loaders
	int is1000 = 0;

	while (argc > 1 && argv[1][0] == '-')
	{
		if (strcmp(argv[1], "-l") == 0)
		{
			outputMedia = TRS80Loader::cassette500;
		}
		else if (strcmp(argv[1], "-h") == 0)
		{
			outputMedia = TRS80Loader::cassette1500;
		}
		else if (strcmp(argv[1], "-t") == 0)
		{
			outputMedia = TRS80Loader::cassette250;
		}
		else if (strcmp(argv[1], "-2") == 0)
		{
			outputMedia = TRS80Loader::cassette500;
			is1000 = 1;
		}
		else if (strcmp(argv[1], "-d") == 0)
		{
			sl.SetDump(true);
			sl.SetDumpData(true);
			sl.m_load.SetCheckChecksum(false);
			dump = 1;
		}
		else if (strcmp(argv[1], "-n") == 0)
		{
			use_name = 1;
		}
		else if (strcmp(argv[1], "-inf") == 0)
		{
			sl.SetDump(true);
			sl.SetDumpData(false);
			sl.m_load.SetCheckChecksum(false);
			dump = 1;
		}
		else if (strcmp(argv[1], "-i") == 0)
		{
			sl.m_load.SetCheckChecksum(false);
		}
		else if (strcmp(argv[1], "-s") == 0)
		{
			argc--;
			argv++;
			if (argc < 2)
			{
				usage();
				return 1;
			}
			stopStage = atoi(argv[1]);
		}
		else if (strcmp(argv[1], "-c") == 0)
		{
			pad = 0;
		}
		else if (strcmp(argv[1], "-v") == 0)
		{
			sl.m_verbose = true;
		}
		else
		{
			usage();
			return 1;
		}

		argc--;
		argv++;
	}

	if (argc < 3 - dump)
	{
		usage();
		return 1;
	}

	// Support our oddity of translating BASIC to SYSTEM.
	sl.m_load.SetBasicEntryAddress(0x1a19);

	for (int i = 1; i < argc - !dump; i++)
	{
		if (dump)
			printf("%s: ", argv[i]);

		sl.m_filename = argv[i];
		sl.m_stage = 0;
		sl.m_stopStage = stopStage;
		sl.SetStopLoad(false);

		if (!sl.LoadFileData(argv[i]))
		{
			printf("trld: error loading '%s': %s\n", argv[i], sl.GetError());
			if (dump)
				continue;

			return 2;
		}

		const char *extension = strrchr(argv[i], '.');

		if (!sl.LoadNext(true, stricmp(extension, ".rom") == 0))
		{
			printf("trld: error loading '%s': %s\n", argv[i], sl.GetError());
			if (dump)
				continue;

			return 3;
		}

		if (!dump)
		{
			if (!sl.m_stopLoad)
			{
				if (!sl.Analyze())
				{
					printf("trld: failed.\n");
					return 4;
				}
			}
		}
	}

	if (dump)
	{
		return 0;
	}

	TRS80Loader::Format format = TRS80Loader::system;

	// Decide output format and media type from output file name.
	for (const char *tail = argv[argc - 1]; *tail; tail++)
	{
		if (stricmp(tail, ".cmd") == 0)
		{
			format = TRS80Loader::cmd;
			outputMedia = TRS80Loader::justBytes;
			break;
		}
		else if (stricmp(tail, ".hex") == 0)
		{
			format = TRS80Loader::hex;
			outputMedia = TRS80Loader::justBytes;
			break;
		}
		else if (stricmp(tail, ".tcas") == 0 ||
				 stricmp(tail, ".250.cas") == 0)
		{
			outputMedia = TRS80Loader::cassette250;
			break;
		}
		else if (stricmp(tail, ".lcas") == 0 ||
				 stricmp(tail, ".500.cas") == 0)
		{
			outputMedia = TRS80Loader::cassette500;
			break;
		}
		else if (stricmp(tail, ".2cas") == 0 ||
				 stricmp(tail, ".1000.cas") == 0)
		{
			// 1000 baud is identical to 500 baud in .cas format.
			outputMedia = TRS80Loader::cassette500;
			break;
		}
		else if (stricmp(tail, ".hcas") == 0 ||
				 stricmp(tail, ".1500.cas") == 0)
		{
			outputMedia = TRS80Loader::cassette1500;
			break;
		}
		else if (stricmp(tail, ".250.wav") == 0)
		{
			outputMedia = TRS80Loader::cassette250;
			outputPhysical = TRS80Loader::baud250;
			break;
		}
		else if (stricmp(tail, ".1000.wav") == 0)
		{
			outputMedia = TRS80Loader::cassette500;
			outputPhysical = TRS80Loader::baud1000;
			break;
		}
		else if (stricmp(tail, ".500.wav") == 0)
		{
			outputMedia = TRS80Loader::cassette500;
			outputPhysical = TRS80Loader::baud500;
			break;
		}
		else if (stricmp(tail, ".1500.wav") == 0)
		{
			outputMedia = TRS80Loader::cassette1500;
			outputPhysical = TRS80Loader::baud1500;
			break;
		}
		else if (stricmp(tail, ".cas") == 0)
		{
			switch (sl.m_load.GetMediaType())
			{
			case TRS80Loader::cassette250:
				outputMedia = TRS80Loader::cassette250;
				break;
			case TRS80Loader::cassette500:
				// 1000 baud is identical to 500 baud in .cas format.
				outputMedia = TRS80Loader::cassette500;
				break;
			default:
			case TRS80Loader::cassette1500:
				outputMedia = TRS80Loader::cassette1500;
				break;
			}
		}
		else if (stricmp(tail, ".wav") == 0)
		{
			switch (sl.m_load.GetMediaType())
			{
			case TRS80Loader::cassette250:
				outputMedia = TRS80Loader::cassette250;
				outputPhysical = TRS80Loader::baud250;
				break;
			case TRS80Loader::cassette500:
				outputMedia = TRS80Loader::cassette500;
				outputPhysical =
					is1000 ? TRS80Loader::baud1000 : TRS80Loader::baud500;
				break;
			default:
			case TRS80Loader::cassette1500:
				outputMedia = TRS80Loader::cassette1500;
				outputPhysical = TRS80Loader::baud1500;
				break;
			}
		}
		else if (stricmp(tail, ".rom") == 0)
		{
			format = TRS80Loader::rom;
			outputMedia = TRS80Loader::justBytes;
			break;
		}
	}

	int low = sl.m_load.GetLowAddress();

	if (format == TRS80Loader::cmd)
	{
		if (low < 0x5200)
			printf("Warning: load at $%04x will overwrite Model I+III, TRS-DOS\n", low);
	}
	else if (format == TRS80Loader::system)
	{
		if (outputMedia == TRS80Loader::cassette250)
		{
			if (low < 0x4091)
				printf("Warning: load at $%04x may interfere with Level I loader\n", low);
		}
		else
		{
			if (low < 0x4288)
				printf("Warning: load at $%04x may interfere with SYSTEM\n", low);
		}
	}

	char name[32];
	NameFromFilename(format, argv[argc - 1], name);

	int parts = use_name ? ~0 : ~TRS80Loader::enc_name;
	if (!pad || outputPhysical != TRS80Loader::justBits)
	{
		parts &= ~TRS80Loader::enc_padding;
	}

	if (sl.m_load.Encode(name, format, outputMedia, parts))
	{
		sl.m_load.SaveFile(argv[argc - 1], outputPhysical);
	}

	char *errorMessage = sl.m_load.GetErrorMessage();

	if (errorMessage)
	{
		printf("trld: error saving '%s': %s\n", argv[argc - 1], errorMessage);
		delete errorMessage;
		return 3;
	}

	return 0;
}

static void usage()
{
	printf("trld version 1.5.2    http://48k.ca/trld.html\n");
	printf("Converts and combines TRS-80 executables to/from various formats.\n");
	printf("\n");
	printf("usage: trld {options} file1.cmd {file2.cas, file3.hex} output.cmd\n");
	printf("\t-c\tClean - don't pad .cas tail with zero bits\n");
	printf("\t-i\tIgnore checksum errors.\n");
	printf("\t-t\t250 baud output (if output is a .cas or .wav file)\n");
	printf("\t-l\t500 baud output (if output is a .cas or .wav file)\n");
	printf("\t-2\t1000 baud output (if output is a .cas or .wav file)\n");
	printf("\t-h\t1500 baud output (if output is a .cas or .wav file)\n");
	printf("\t\t-n\tPut file name in .cmd output\n");
	printf("\t-s N\tStop at stage N of relocation and/or cassette loading.\n");
	printf("\t\t'-s 1' to do nothing, '-s 0' to keep trying indefinitely.\n");
	printf("\t-v\tOutput extra information for debugging purposes.\n");
	printf("\n");
	printf("The output format will change depending on the output file name's suffix:\n");
	printf("\t.cmd\t\t\tTRS-DOS program data\n");
	printf("\t.hex\t\t\tIntel HEX record data\n");
	printf("\t.cas\t\t\tcassette format, 1500 baud or input's baud rate\n");
	printf("\t.tcas .250.cas\t\t250 baud cassette Level 1 CLOAD data\n");
	printf("\t.lcas .500.cas\t\t500 baud cassette SYSTEM data\n");
	printf("\t.2cas .1000.cas\t\t1000 baud cassette SYSTEM data\n");
	printf("\t.hcas .1500.cas\t\t1500 baud cassette SYSTEM data\n");
	printf("\t.wav\t\t\tcassette audio, 1500 baud or input's baud rate\n");
	printf("\t.1500.wav\t\t1500 baud cassette SYSTEM audio\n");
	printf("\t.1000.wav\t\t1000 baud cassette SYSTEM audio\n");
	printf("\t.500.wav\t\t500 baud cassette SYSTEM audio\n");
	printf("\t.250.wav\t\t250 baud cassette Level 1 CLOAD audio\n");
	printf("\n");
	printf("or: trld -d {file}\n");
	printf("\tdumps load records from one or more files\n");
	printf("or: trld -inf {file}\n");
	printf("\tgives type on one or more files\n");
}
